/* wmcdclock - Josef Kufner <jkufner@seznam.cz>
 * Dockapp Clock
 *
 * based on WMCPU 1.3 by timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>


#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "master.xpm"

#define WMCDCLOCK_VERSION "0.1"

#define FIFO_BUF_MAX	128

#define copy_xpm_area(x, y, w, h, dx, dy) \
    XCopyArea(display, pixmap, pixmap, gc, x, y, w, h, dx, dy)
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

#define HELP_TEXT \
	"wmcdclock: clock dockapp that looks like xosview\n" \
	"\tJosef Kufner <jkufner@seznam.cz>\n" \
	"usage:\n" \
	"\t-display <display name>\n" \
	"\t-h\tthis screen\n" \
	"\t-v\tprint the version number\n"

static char *display_name = NULL;
static Display *display;
static Window win;
static Window iconwin;
static GC gc;
static Pixmap pixmap;
static Pixmap mask;

static int count_down = 0;
static time_t count_down_time = 0;
static char *fifo_name = NULL;
static int fifo_fd = 0;

static int done = 0;

static void wmcdclock_routine(void);
static void redraw_window(void);
static void new_window(char *name, int width, int height);
static void on_signal(int sig);
static void on_end (void);
static void open_fifo(void);
static void close_fifo(void);
static void parse_fifo(char *buf);


int main(int argc, char **argv) {
	int ch;
	
	tzset();
	
	while ((ch = getopt(argc, argv, "d:hvf:")) != EOF) {
		switch (ch) {
			case 'd':
				if (optarg) display_name = strdup(optarg);
				break;
			case 'v':
				fprintf(stderr, "%s\n", WMCDCLOCK_VERSION);
				exit(0);
				break;
			case 'f':
				if (optarg) fifo_name = strdup(optarg);
				break;
			default:
				fputs(HELP_TEXT, stdout);
				exit(0);
				break;
		}
	}

	if (signal(SIGINT, on_signal) == SIG_ERR
			|| signal(SIGTERM, on_signal) == SIG_ERR
			|| signal(SIGQUIT, on_signal) == SIG_ERR
			|| signal(SIGPIPE, SIG_IGN) == SIG_ERR) 
	{
		perror("wmcdclock: signal");
		return EXIT_FAILURE;
	}
	
	if ((display = XOpenDisplay(display_name)) == NULL) {
		fprintf(stderr, "Unable to open display \"%s\"\n", display_name);
		return EXIT_FAILURE;
	}
	
	if (fifo_name) {
		if (mkfifo(fifo_name, 0600) != 0) {
			perror("wmcdclock: mkfifo");
			exit(EXIT_FAILURE);
		}
		open_fifo();
	}
	
	atexit(on_end);
	
	wmcdclock_routine();
	
	return EXIT_SUCCESS;
}


void wmcdclock_routine(void) {
	XEvent Event;
	time_t time_new = 1, time_old = 0, time_show;
	struct tm now, cd;
	fd_set inputs;
	int xfd = 0, zero_blink = 0;
	unsigned int loop;
	struct timeval timeout;
	char buf[32], fifo_buf[FIFO_BUF_MAX];
	size_t rd, fifo_buf_size = 0;

	new_window("wmcdclock", 64,  64);

	timeout.tv_sec = 0;
	timeout.tv_usec = 150000;

	strcpy(fifo_buf, "");

	for (loop = 0; !done; loop++) {
		time_new = time(NULL);
		if (time_new != time_old) {
			/* update window */
			now = *localtime(&time_new);
			/* time */
			copy_xpm_area(0 + now.tm_hour / 10 * 12, 68, 12, 16,  7, 7);
			copy_xpm_area(0 + now.tm_hour % 10 * 12, 68, 12, 16, 19, 7);
			copy_xpm_area(70 + (now.tm_sec < 30 ? 5 : 10) * (time_new % 2), 0, 3, 16, 30, 7);
			copy_xpm_area(0 + now.tm_min  / 10 * 12, 68, 12, 16, 35, 7);
			copy_xpm_area(0 + now.tm_min  % 10 * 12, 68, 12, 16, 47, 7);

			/* week day */
			copy_xpm_area(0 + 13 * now.tm_wday, 95, 13, 7, 6, 33);
			
			/* date */
			copy_xpm_area(0 + now.tm_mday / 10 * 6, 86, 6, 7, 29, 33);
			copy_xpm_area(0 + now.tm_mday % 10 * 6, 86, 6, 7, 35, 33);
			copy_xpm_area(0 + (now.tm_mon+1) / 10 * 6, 86, 6, 7, 43, 33);
			copy_xpm_area(0 + (now.tm_mon+1) % 10 * 6, 86, 6, 7, 49, 33);
			
			/* count down */
			if (count_down) {
				time_show = count_down_time - time_new;
				cd = *gmtime(&time_show);
				copy_xpm_area(0 + cd.tm_hour / 10 * 6, 86, 6, 7, 10, 49);
				copy_xpm_area(0 + cd.tm_hour % 10 * 6, 86, 6, 7, 16, 49);
				copy_xpm_area(105, 0, 1, 6, 23, 49);
				copy_xpm_area(0 + cd.tm_min / 10 * 6, 86, 6, 7, 26, 49);
				copy_xpm_area(0 + cd.tm_min % 10 * 6, 86, 6, 7, 32, 49);
				copy_xpm_area(105, 0, 1, 6, 39, 49);
				copy_xpm_area(0 + cd.tm_sec / 10 * 6, 86, 6, 7, 42, 49);
				copy_xpm_area(0 + cd.tm_sec % 10 * 6, 86, 6, 7, 48, 49);

				if (count_down_time <= time_new) {
					count_down = 0;
					zero_blink = 41 + (loop % 2);
				}
			} else {
				copy_xpm_area(76, 48, 43, 8, 10, 48);
			}

			redraw_window();
			time_old = time_new;
		}

		if (count_down == 0 && zero_blink > 0) {
			zero_blink--;
			if (loop % 2) {
				copy_xpm_area(0, 86, 6, 7, 10, 49);
				copy_xpm_area(0, 86, 6, 7, 16, 49);
				copy_xpm_area(105, 0, 1, 6, 23, 49);
				copy_xpm_area(0, 86, 6, 7, 26, 49);
				copy_xpm_area(0, 86, 6, 7, 32, 49);
				copy_xpm_area(105, 0, 1, 6, 39, 49);
				copy_xpm_area(0, 86, 6, 7, 42, 49);
				copy_xpm_area(0, 86, 6, 7, 48, 49);
			} else {
				copy_xpm_area(76, 48, 43, 8, 10, 48);
			}
			redraw_window();
		}
		
		xfd = ConnectionNumber(display);
		FD_ZERO(&inputs);
		FD_SET(xfd, &inputs);
		if (fifo_fd)
			FD_SET(fifo_fd, &inputs);
		
		switch (select(FD_SETSIZE, &inputs, NULL, NULL, &timeout)) {
			case 0:
				timeout.tv_sec = 0;
				timeout.tv_usec = 150000;
				continue;
			case -1:
				continue;
			default:
				break;
				
		}

		if (FD_ISSET(xfd, &inputs) && XPending(display)) {
			XNextEvent(display, &Event);
			switch (Event.type) {
				case Expose:
					redraw_window();
					break;
				case DestroyNotify:
					XCloseDisplay(display);
					done = 1;
					continue;
			}
		} else if (FD_ISSET(fifo_fd, &inputs)) {
			if ((rd = read(fifo_fd, buf, 31)) == 0 || rd + fifo_buf_size > FIFO_BUF_MAX) {
				close_fifo();
				open_fifo();
				parse_fifo(fifo_buf);
				fifo_buf_size = 0;
				strcpy(fifo_buf, "");
			} else {
				buf[rd] = '\0';
				strcat(fifo_buf, buf);
				fifo_buf_size += rd;
			}
		}
	}
}


static void on_signal (int sig) {
	if (sig == SIGINT) {
		done = 1;
	} else {
		fprintf(stderr, "opss... jestli se tohle stalo, tak je neco v haji...\n");
	}
	return;
}

static void on_end (void) {
	if (fifo_name) {
		close_fifo();
		if (unlink(fifo_name) != 0) {
			perror("wmcdclock: unlink");
			exit(EXIT_FAILURE);
		}
	}
}

static void open_fifo(void) {
	if ((fifo_fd = open(fifo_name, O_RDONLY | O_NONBLOCK)) == -1) {
		perror("wmcdclock: open fifo");
		exit(EXIT_FAILURE);
	}
}

static void close_fifo(void) {
	if (fifo_fd && close(fifo_fd) == -1) {
		perror("wmcdclock: close fifo");
		exit(EXIT_FAILURE);
	}
}


static void parse_fifo(char *buf) {
	int sec, min, hour;
	char cmd;
	time_t set_ts;
	struct tm set;

	if (sscanf(buf, "%c%02d:%02d:%02d", &cmd, &hour, &min, &sec) != 4 && cmd != '+' && cmd != '=') {
		fprintf(stderr, "wmcdclock: wrong command -- must be [+=]HH:MM:SS\n");
		return;
	}

	set_ts = time(NULL);
	if (cmd == '=') {
		set = *localtime(&set_ts);
		set.tm_mday += (hour < set.tm_hour);
		set.tm_hour = hour;
		set.tm_min = min;
		set.tm_sec = sec;
	} else if (cmd == '+') {
		set = *localtime(&set_ts);
		set.tm_hour += hour;
		set.tm_min += min;
		set.tm_sec += sec;
	}
	count_down_time = mktime(&set);
	count_down = 1;
}


static void new_window(char *name, int width, int height) {
    XpmAttributes attr;
    Pixel fg, bg;
    XGCValues gcval;
    XSizeHints sizehints;
    XClassHint classhint;
    XWMHints wmhints;
    XTextProperty wname;

    sizehints.flags = USSize | USPosition;
    sizehints.x = 0;
    sizehints.y = 0;
    sizehints.width = width;
    sizehints.height = height;

    fg = BlackPixel(display, DefaultScreen(display));
    bg = WhitePixel(display, DefaultScreen(display));

    win = XCreateSimpleWindow(display, DefaultRootWindow(display),
			      sizehints.x, sizehints.y,
			      sizehints.width, sizehints.height, 1, fg,
			      bg);

    iconwin = XCreateSimpleWindow(display, win, sizehints.x, sizehints.y,
				  sizehints.width, sizehints.height, 1, fg,
				  bg);

    XSetWMNormalHints(display, win, &sizehints);
    classhint.res_name = name;
    classhint.res_class = name;
    XSetClassHint(display, win, &classhint);

#define INPUT_MASK \
    ButtonPressMask \
    | ExposureMask

    XSelectInput(display, win, INPUT_MASK);
    XSelectInput(display, iconwin, INPUT_MASK);

#undef INPUT_MASk

    XStringListToTextProperty(&name, 1, &wname);
    XSetWMName(display, win, &wname);

    gcval.foreground = fg;
    gcval.background = bg;
    gcval.graphics_exposures = 0;

    gc = XCreateGC(display, win,
	    GCForeground | GCBackground | GCGraphicsExposures,
	    &gcval);

    attr.exactColors = 0;
    attr.alloc_close_colors = 1;
    attr.closeness = 30000;
    attr.valuemask = (XpmExactColors | XpmAllocCloseColors | XpmCloseness);
    if ((XpmCreatePixmapFromData(display, DefaultRootWindow(display),
				master_xpm, &pixmap, &mask,
				&attr) != XpmSuccess)) {
	fputs("Cannot allocate colors for the dockapp pixmap!\n", stderr);
	exit(EXIT_FAILURE);
    }

    XShapeCombineMask(display, win, ShapeBounding, 0, 0, mask,
		      ShapeSet);
    XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, mask,
		      ShapeSet);

    wmhints.initial_state = WithdrawnState;
    wmhints.icon_window = iconwin;
    wmhints.icon_x = sizehints.x;
    wmhints.icon_y = sizehints.y;
    wmhints.window_group = win;
    wmhints.flags =
	StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    XSetWMHints(display, win, &wmhints);

    XMapWindow(display, win);
}


static void redraw_window(void) {
	XCopyArea(display, pixmap, iconwin, gc, 0, 0, 64, 64, 0, 0);
	XCopyArea(display, pixmap, win, gc, 0, 0, 64, 64, 0, 0);
	XFlush(display);
}
